@@ -2,7 +2,7 @@ import copy
 import json
 from collections import OrderedDict
 from types import FunctionType
-from typing import Type, Dict, Tuple, Sequence
+from typing import Type, Dict, Tuple, Sequence, Callable
 
 import inflection
 from django.db.models import QuerySet, Model
@@ -12,6 +12,7 @@ from django_filters.rest_framework import FilterSet
 from django_filters.utils import get_model_field
 from rest_framework.fields import get_attribute
 from rest_framework.relations import ManyRelatedField, MANY_RELATION_KWARGS
+from rest_framework.utils.serializer_helpers import ReturnDict
 from rest_framework_json_api import serializers
 from rest_framework_json_api.django_filters import DjangoFilterBackend
 from rest_framework_json_api.relations import ResourceRelatedField
@@ -19,14 +20,15 @@ from rest_framework_json_api.utils import get_resource_type_from_serializer, get
     get_resource_type_from_queryset
 
 from .generic_relation import GenericRelatedField
-from .namespace import _RESOURCE_NAME_TO_SPICE, _MODEL_TO_SERIALIZER
-from .types import CustomField, Relation, Filter, GenericRelation
+from .namespace import _RESOURCE_NAME_TO_SPICE, _MODEL_TO_SERIALIZERS
+from .types import CustomField, Relation, Filter, GenericRelation, ComputedFilter
 
 
 def _construct_serializer(serializer_prefix: str, model: Type[Model], resource_name: str, fields: Sequence[str],
                           custom_fields: Sequence[CustomField], relations: Sequence[Relation],
                           generic_relations: Sequence[GenericRelation], related_limit: int,
-                          primary_key_name: str, on_validate: FunctionType = None) -> Type:
+                          primary_key_name: str, on_validate: FunctionType = None,
+                          after_list_callback: Callable = None) -> Type:
     def to_representation(self, iterable):
         if isinstance(iterable, QuerySet):
             real_count = iterable.count()
@@ -196,12 +198,19 @@ def _construct_serializer(serializer_prefix: str, model: Type[Model], resource_n
     class GenericSerializer(serializers.HyperlinkedModelSerializer):
         def __new__(cls, instance=None, *args, **kwargs):
             check_instance = instance
-            if isinstance(instance, (list,)) and len(instance) > 0:
+            if isinstance(instance, (list, QuerySet)) and len(instance) > 0:
                 check_instance = instance[0]
-            if check_instance is not None and not isinstance(check_instance, (cls.Meta.model, list,)):
-                return _MODEL_TO_SERIALIZER[type(check_instance)](instance=instance, *args, **kwargs)
+            if check_instance is not None and not isinstance(check_instance, (cls.Meta.model, list, QuerySet)):
+                return _MODEL_TO_SERIALIZERS[type(check_instance)][0](instance=instance, *args, **kwargs)
             return super(GenericSerializer, cls).__new__(cls, instance=instance, *args, **kwargs)
 
+        @property
+        def data(self):
+            ret = super().data
+            if after_list_callback is not None:
+                ret = after_list_callback({'results': [ret]})['results'][0]
+            return ReturnDict(ret, serializer=self)
+
     new_serializer = type(f'{serializer_prefix}{resource_name}Serializer', (GenericSerializer,), {
         **{custom_field.name: serializers.SerializerMethodField(read_only=True) for custom_field in
            custom_fields},
@@ -216,7 +225,7 @@ def _construct_serializer(serializer_prefix: str, model: Type[Model], resource_n
             related_link_lookup_field=primary_key_name,
             related_link_url_kwarg=relation.primary_key_name or 'id',
             self_link_view_name=f'{resource_name}-relationships'
-        ) for relation in relations},
+        ) for relation in relations if hasattr(model, relation.field)},
         **{relation.field: GenericRelatedField(
             {
                 related_model: generate_generic_resource(related_model)(
@@ -233,19 +242,24 @@ def _construct_serializer(serializer_prefix: str, model: Type[Model], resource_n
                 for related_model, relation_resource_name in getattr(relation, 'related', {}).items()
             },
             self_link_view_name=f'{resource_name}-relationships',
-            related_link_lookup_field=primary_key_name) for relation in generic_relations},
+            related_link_lookup_field=primary_key_name) for relation in generic_relations if
+            hasattr(model, relation.field)},
         'validate': validate_data,
-        'Meta': type('Meta', (), {'model': model, 'fields': [*fields, *list(
-            map(lambda custom_field: custom_field.name, custom_fields))],
-                                  'resource_name': resource_name}),
+        'Meta': type('Meta', (),
+                     {'model': model, 'fields': [*[field for field in [*fields] if hasattr(model, field)], *list(
+                         map(lambda custom_field: custom_field.name, custom_fields))],
+                      'resource_name': resource_name}),
         'included_serializers': included_serializers,
         'included_generic_serializers': included_generic_serializers
     })
-    _MODEL_TO_SERIALIZER[model] = new_serializer
+    if model not in _MODEL_TO_SERIALIZERS:
+        _MODEL_TO_SERIALIZERS[model] = []
+    _MODEL_TO_SERIALIZERS[model].append(new_serializer)
     return new_serializer
 
 
-def _construct_filter_backend(model: Type[Model], resource_name: str, filters: Dict[str, Filter]) -> Tuple[Type, Type]:
+def _construct_filter_backend(model: Type[Model], resource_name: str, filters: Dict[str, Filter],
+                              computed_filters: Dict[str, ComputedFilter]) -> Tuple[Type, Type]:
     constructed_filters_transform_callbacks = {}
     constructed_filters = {}
     for key, filter in filters.items():
@@ -261,6 +275,10 @@ def _construct_filter_backend(model: Type[Model], resource_name: str, filters: D
 
     filter_set = type(f'{resource_name}FilterSet', (FilterSet,), {
         **constructed_filters,
+        **{field: computed_filter.filter_type(method=f'filter_{field}') for field, computed_filter in
+           computed_filters.items()},
+        **{f'filter_{field}': lambda self, queryset, field_name, value: computed_filter.filter_func(queryset, value) for
+           field, computed_filter in computed_filters.items()},
         'Meta': type('Meta', (), {
             'model': model,
             'fields': []
