@@ -1,7 +1,7 @@
 from copy import deepcopy
 from functools import partial
 from types import FunctionType
-from typing import Type, Tuple, Sequence, Dict, Callable, Any
+from typing import Type, Tuple, Sequence, Dict, Callable, Any, Optional
 
 from django.apps import apps
 from django.conf.urls import url
@@ -24,24 +24,27 @@ from . import lookups as filter_lookups
 from . import plugins
 from .common import LimitedJsonApiPageNumberPagination, JsonApiSearchFilter, LOGGER
 from .constructors import _construct_serializer, _construct_filter_backend
-from .namespace import _append_to_namespace, _RESOURCE_NAME_TO_SPICE
-from .types import CustomField, Filter, Relation, GenericRelation
+from .namespace import _append_to_namespace, _RESOURCE_NAME_TO_SPICE, _MODEL_TO_SERIALIZERS
+from .types import CustomField, Filter, Relation, GenericRelation, ComputedFilter
 
 
 class JsonApiModelViewBuilder:
     DEFAULT_RELATED_LIMIT = 100
 
-    def __init__(self, model: Type[Model], primary_key_name: str = None, resource_name: str = None,
+    def __init__(self, model: Type[Model], primary_key_name: Optional[str] = None, resource_name: Optional[str] = None,
                  allowed_methods=json_api_spec_http_methods.HTTP_ALL,
-                 permission_classes: Sequence[Type[BasePermission]] = None,
-                 authentication_classes: Sequence[Type[BaseAuthentication]] = None,
-                 queryset: QuerySet = None,
-                 permitted_objects: Callable[[Request, QuerySet], QuerySet] = None,
-                 skip_plugins: Sequence[str] = None):
+                 permission_classes: Optional[Sequence[Type[BasePermission]]] = None,
+                 authentication_classes: Optional[Sequence[Type[BaseAuthentication]]] = None,
+                 queryset: Optional[QuerySet] = None,
+                 permitted_objects: Optional[Callable[[Request, QuerySet], QuerySet]] = None,
+                 skip_plugins: Optional[Sequence[str]] = None,
+                 include_plugins: Optional[Sequence[str]] = None,
+                 plugin_options: Optional[Dict[str, Any]] = None):
         self.__validate_http_methods(allowed_methods)
         self._model = model
         self._fields = {}
         self._filters = {}
+        self._computed_filters = {}
         self._relations = {}
         self._generic_relations = {}
         self._custom_fields = {}
@@ -51,18 +54,25 @@ class JsonApiModelViewBuilder:
         self._related_limit = self.DEFAULT_RELATED_LIMIT
         self._permission_classes = permission_classes or []
         self._authentication_classes = authentication_classes or []
-        self._pre_create_callback = None
-        self._post_create_callback = None
-        self._pre_update_callback = None
-        self._post_update_callback = None
-        self._pre_delete_callback = None
-        self._post_delete_callback = None
+        self._before_create_callback = None
+        self._after_create_callback = None
+        self._after_get_callback = None
+        self._before_update_callback = None
+        self._after_update_callback = None
+        self._before_delete_callback = None
+        self._after_delete_callback = None
+        self._after_list_callback = None
         if queryset is None:
             self._queryset = self._model.objects
         else:
             self._queryset = queryset
         self._spice_queryset = permitted_objects
-        self._skip_plugins = skip_plugins or []
+        self._skip_plugins = skip_plugins if skip_plugins is not None else [plugins.AUTO_ADMIN_VIEWS]
+        include_plugins = include_plugins or []
+        for item in include_plugins:
+            if item in self._skip_plugins:
+                self._skip_plugins.remove(item)
+        self._plugin_options = plugin_options or {}
 
     @staticmethod
     def __validate_http_methods(limit_to_http_methods: Sequence[str] = json_api_spec_http_methods.HTTP_ALL):
@@ -84,12 +94,23 @@ class JsonApiModelViewBuilder:
         self._fields[limit_to_on_retrieve].extend(fields)
         return self
 
+    def dummy_fields(self, fields: Sequence[str]) -> 'JsonApiModelViewBuilder':
+        self.fields(fields=fields)
+        _f = lambda instance: ''
+        self.custom_fields(fields=[(field, _f,) for field in fields])
+        return self
+
     def add_field(self, name: str, limit_to_on_retrieve: bool = False) -> 'JsonApiModelViewBuilder':
         if limit_to_on_retrieve not in self._fields:
             self._fields[limit_to_on_retrieve] = []
         self._fields[limit_to_on_retrieve].append(name)
         return self
 
+    def add_dummy_field(self, name: str) -> 'JsonApiModelViewBuilder':
+        self.add_field(name=name)
+        self.add_custom_field(name=name, instance_callback=lambda instance: '')
+        return self
+
     def add_filter(self, name: str, field: str = None, lookups: Sequence[str] = None,
                    transform_value: Callable[
                        [str, QuerySet], Tuple[str, QuerySet]] = None) -> 'JsonApiModelViewBuilder':
@@ -102,6 +123,13 @@ class JsonApiModelViewBuilder:
         self._filters[name] = Filter(field=field or name, lookups=lookups, transform_value=transform_value)
         return self
 
+    def add_computed_filter(self, name: str, filter_type: Filter,
+                            filter_func: Callable[[QuerySet], str],
+                            field: str = None) -> 'JsonApiModelViewBuilder':
+        self._computed_filters[name] = ComputedFilter(field=field or name, filter_func=filter_func,
+                                                      filter_type=filter_type)
+        return self
+
     def add_relation(self, field: str, many: bool = False, resource_name: str = None,
                      primary_key_name: str = None,
                      limit_to_on_retrieve: bool = False,
@@ -150,44 +178,55 @@ class JsonApiModelViewBuilder:
         self._related_limit = limit
         return self
 
-    def pre_create(self, pre_create_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._pre_create_callback = pre_create_callback
+    def before_create(self, before_create_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._before_create_callback = before_create_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_POST)
         return self
 
-    def post_create(self, pre_create_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._post_create_callback = pre_create_callback
+    def after_create(self, after_create_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._after_create_callback = after_create_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_POST)
         return self
 
-    def pre_update(self, pre_update_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._pre_update_callback = pre_update_callback
+    def after_get(self, after_get_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._after_get_callback = after_get_callback
+        self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
+        return self
+
+    def before_update(self, before_update_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._before_update_callback = before_update_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_PATCH)
         return self
 
-    def post_update(self, pre_update_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._post_update_callback = pre_update_callback
+    def after_update(self, after_update_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._after_update_callback = after_update_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_PATCH)
         return self
 
-    def pre_delete(self, pre_delete_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._pre_delete_callback = pre_delete_callback
+    def before_delete(self, before_delete_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._before_delete_callback = before_delete_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_DELETE)
         return self
 
-    def post_delete(self, pre_delete_callback: FunctionType = None) -> 'JsonApiModelViewBuilder':
-        self._post_delete_callback = pre_delete_callback
+    def after_delete(self, after_delete_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._after_delete_callback = after_delete_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_DELETE)
         return self
 
+    def after_list(self, after_list_callback: Callable[[Any], Any] = None) -> 'JsonApiModelViewBuilder':
+        self._after_list_callback = after_list_callback
+        self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
+        return self
+
     def _get_history_urls(self) -> Sequence[partial]:
         history_builder = deepcopy(self)
-        history_builder._skip_plugins = [plugins.DJANGO_SIMPLE_HISTORY]
+        history_builder._skip_plugins = [plugins.DJANGO_SIMPLE_HISTORY, plugins.AUTO_ADMIN_VIEWS]
         history_builder._model = apps.get_model(self._model.objects.model._meta.db_table.split('_')[0],
                                                 f'Historical{self._model.__name__}')
         history_builder._queryset = self._model.history
         history_builder._resource_name = f'historical_{self._resource_name}'
 
+        history_builder._custom_fields = []
         history_urls = history_builder.fields(['history_date', 'history_change_reason', 'history_id', 'history_type']) \
             .add_filter(name='history_date', lookups=(
             filter_lookups.EXACT, filter_lookups.IN, filter_lookups.LT, filter_lookups.LTE, filter_lookups.GT,
@@ -199,53 +238,89 @@ class JsonApiModelViewBuilder:
 
         return history_urls
 
-    def _build(self, url_resource_name: str = '', urls_prefix: str = '') -> Sequence[partial]:
+    def _get_admin_urls(self) -> Sequence[partial]:
+        admin_builder = deepcopy(self)
+        admin_builder._skip_plugins = [plugins.AUTO_ADMIN_VIEWS]
+        admin_builder._spice_queryset = None
+        admin_builder._resource_name = f'admin_view_{admin_builder._resource_name}'
+        admin_permission_class = admin_builder._plugin_options.get(plugins.AUTO_ADMIN_VIEWS, {}).get(
+            'ADMIN_PERMISSION_CLASS')
+
+        if admin_permission_class is not None:
+            admin_builder._permission_classes = [*admin_builder._permission_classes, admin_permission_class]
+
+        admin_urls = admin_builder._build(url_resource_name=self._resource_name, urls_prefix='admin/',
+                                          ignore_serializer=True)
+
+        return admin_urls
+
+    def _build(self, url_resource_name: str = '', urls_prefix: str = '', ignore_serializer: bool = False) -> Sequence[
+        partial]:
         method_to_serializer = {}
-        for limit_to_on_retrieve in [False, True]:
-            fields = self._fields[limit_to_on_retrieve] if limit_to_on_retrieve in self._fields else []
-            if limit_to_on_retrieve is True:
-                fields.extend(self._fields[False] if False in self._fields else [])
-            custom_fields = self._custom_fields[
-                limit_to_on_retrieve] if limit_to_on_retrieve in self._custom_fields else []
-            if limit_to_on_retrieve is True:
-                custom_fields.extend(self._custom_fields[False] if False in self._custom_fields else [])
-            relations = self._relations[limit_to_on_retrieve] if limit_to_on_retrieve in self._relations else []
-            if limit_to_on_retrieve is True:
-                relations.extend(self._relations[False] if False in self._relations else [])
-            generic_relations = self._generic_relations[
-                limit_to_on_retrieve] if limit_to_on_retrieve in self._generic_relations else []
-            if limit_to_on_retrieve is True:
-                generic_relations.extend(self._generic_relations[False] if False in self._generic_relations else [])
+        if not ignore_serializer:
+            for limit_to_on_retrieve in [False, True]:
+                fields = self._fields[limit_to_on_retrieve] if limit_to_on_retrieve in self._fields else []
+                if limit_to_on_retrieve is True:
+                    fields.extend(self._fields[False] if False in self._fields else [])
+                custom_fields = self._custom_fields[
+                    limit_to_on_retrieve] if limit_to_on_retrieve in self._custom_fields else []
+                if limit_to_on_retrieve is True:
+                    custom_fields.extend(self._custom_fields[False] if False in self._custom_fields else [])
+                relations = self._relations[limit_to_on_retrieve] if limit_to_on_retrieve in self._relations else []
+                if limit_to_on_retrieve is True:
+                    relations.extend(self._relations[False] if False in self._relations else [])
+                generic_relations = self._generic_relations[
+                    limit_to_on_retrieve] if limit_to_on_retrieve in self._generic_relations else []
+                if limit_to_on_retrieve is True:
+                    generic_relations.extend(self._generic_relations[False] if False in self._generic_relations else [])
 
-            method_to_serializer[limit_to_on_retrieve] = \
-                _construct_serializer('Retrieve' if limit_to_on_retrieve else 'List', self._model, self._resource_name,
-                                      fields,
-                                      custom_fields,
-                                      relations,
-                                      generic_relations,
-                                      self._related_limit,
-                                      self._primary_key_name,
-                                      self._pre_update_callback if limit_to_on_retrieve else self._pre_create_callback)
-            _append_to_namespace(method_to_serializer[limit_to_on_retrieve])
+                method_to_serializer[limit_to_on_retrieve] = \
+                    _construct_serializer('Retrieve' if limit_to_on_retrieve else 'List', self._model,
+                                          self._resource_name,
+                                          fields,
+                                          custom_fields,
+                                          relations,
+                                          generic_relations,
+                                          self._related_limit,
+                                          self._primary_key_name,
+                                          self._before_update_callback if limit_to_on_retrieve else self._before_create_callback,
+                                          self._after_list_callback)
+                _append_to_namespace(method_to_serializer[limit_to_on_retrieve])
+        else:
+            method_to_serializer[False] = _MODEL_TO_SERIALIZERS[self._model][0]
+            method_to_serializer[True] = _MODEL_TO_SERIALIZERS[self._model][1]
 
-        filter_set, filter_backend = _construct_filter_backend(self._model, self._resource_name, self._filters)
+        filter_set, filter_backend = _construct_filter_backend(self._model, self._resource_name, self._filters,
+                                                               self._computed_filters)
 
         def perform_create(view, serializer):
             instance = serializer.save()
-            if self._post_create_callback is not None:
-                self._post_create_callback(instance, serializer)
+            if self._after_create_callback is not None:
+                self._after_create_callback(instance, serializer)
 
         def perform_destroy(view, instance):
-            if self._pre_delete_callback is not None:
-                self._pre_delete_callback(instance, serializer)
+            if self._before_delete_callback is not None:
+                self._before_delete_callback(instance, view.get_serializer())
             instance.delete()
-            if self._post_delete_callback is not None:
-                self._post_delete_callback(instance, serializer)
+            if self._after_delete_callback is not None:
+                self._after_delete_callback(instance, view.get_serializer())
+
+        def perform_get(view, instance, *args, **kwargs):
+            response = super(view.__class__, view).retrieve(instance, *args, **kwargs)
+            if self._after_get_callback is not None:
+                response.data = self._after_get_callback(response.data)
+            return response
 
         def perform_update(view, serializer):
             instance = serializer.save()
-            if self._post_update_callback is not None:
-                self._post_update_callback(instance, serializer)
+            if self._after_update_callback is not None:
+                self._after_update_callback(instance, serializer)
+
+        def perform_list(view, *args, **kwargs):
+            response = super(view.__class__, view).list(*args, **kwargs)
+            if self._after_list_callback is not None:
+                response.data = self._after_list_callback(response.data)
+            return response
 
         base_model_view_set = type(f'{self._resource_name}JSONApiModelViewSet', (ModelViewSet,), {
             'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
@@ -289,7 +364,8 @@ class JsonApiModelViewBuilder:
                 'filterset_class': filter_set,
                 'lookup_field': pk_name,
                 'perform_create': perform_create,
-                'name': f'list {self._resource_name}'
+                'name': f'list {self._resource_name}',
+                'list': perform_list
             })
 
             get_method_view_set = type(f'Get{self._resource_name}ViewSet', (base_model_view_set,), {
@@ -306,6 +382,7 @@ class JsonApiModelViewBuilder:
                 'lookup_field': pk_name,
                 'perform_update': perform_update,
                 'perform_destroy': perform_destroy,
+                'retrieve': perform_get
             })
 
             if len(urls_prefix) > 0 and urls_prefix[-1] != '/':
@@ -336,9 +413,12 @@ class JsonApiModelViewBuilder:
                 import simple_history
 
                 urls.extend(self._get_history_urls())
-            except:
+            except Exception as e:
                 pass
 
+        if plugins.AUTO_ADMIN_VIEWS not in self._skip_plugins:
+            urls.extend(self._get_admin_urls())
+
         return urls
 
     def get_urls(self, url_resource_name: str = '', urls_prefix: str = '') -> Sequence[partial]:
@@ -464,4 +544,3 @@ def json_api_action_view(action_name: str,
 
 # Backwards Compatibility Support
 JsonApiViewBuilder = JsonApiModelViewBuilder
-
